module("L_EnOceanGateway1", package.seeall)

local bit = require("nixio").bit

------------------------------------------------------------------------------------------------------------------------
-- Flags
------------------------------------------------------------------------------------------------------------------------

local F_DEBUG_MODE          = true
local F_LEARN_MODE          = false
local F_WAIT_FOR_RESPONSE   = false -- If this is 'true', no further messages will be sent
local F_SENDING_IN_PROGRESS = false

------------------------------------------------------------------------------------------------------------------------
-- Constants
------------------------------------------------------------------------------------------------------------------------

-- Services
local SID = {
	ENOCEAN  = "urn:micasaverde-com:serviceId:EnOceanGateway1",
	SW_POWER = "urn:upnp-org:serviceId:SwitchPower1",
	SECURITY = "urn:micasaverde-com:serviceId:SecuritySensor1",
	TEMP     = "urn:upnp-org:serviceId:TemperatureSensor1",
	HUMIDITY = "urn:micasaverde-com:serviceId:HumiditySensor1",
	LIGHT    = "urn:micasaverde-com:serviceId:LightSensor1"
}

local DEVICE_TYPES = {
	BINARY_LIGHT       = "urn:schemas-upnp-org:device:BinaryLight:1",
	DOOR_SENSOR        = "urn:schemas-micasaverde-com:device:DoorSensor:1",
	TEMPERATURE_SENSOR = "urn:schemas-micasaverde-com:device:TemperatureSensor:1",
	HUMIDITY_SENSOR    = "urn:schemas-micasaverde-com:device:HumiditySensor:1",
	LIGHT_SENSOR       = "urn:schemas-micasaverde-com:device:LightSensor:1",
	MOTION_SENSOR      = "urn:schemas-micasaverde-com:device:MotionSensor:1"
}

local DEVICE_FILES = {
	BINARY_LIGHT       = "D_BinaryLight1.xml",
	DOOR_SENSOR        = "D_DoorSensor1.xml",
	TEMPERATURE_SENSOR = "D_TemperatureSensor1.xml",
	HUMIDITY_SENSOR    = "D_HumiditySensor1.xml",
	LIGHT_SENSOR       = "D_LightSensor1.xml",
	MOTION_SENSOR      = "D_MotionSensor1.xml"
}

-- Message types
local SYS_MESSAGE_TYPES = {
	BUSY    =  1,
	ERROR   =  2,
	SUCCESS =  4
}

-- Delays for luup.delay, in seconds
local DELAYS = {
	SEND_NEXT_MESSAGE = 2
}

------------------------------------------------------------------------------------------------------------------------

local SYNC_BYTE = 0x55

local STATE = {
	GET_SYNC_STATE    = 0,
	GET_HEADER_STATE  = 1,
	CHECK_CRC8H_STATE = 2,
	GET_DATA_STATE    = 3,
	CHECK_CRC8D_STATE = 4
}

local PACKET_TYPES = {
	RADIO          = 0x1,
	RESPONSE       = 0x2,
	EVENT          = 0x4,
	COMMON_COMMAND = 0x5
}

local TELEGRAM_TYPES = {
	["RPS"] = 0xF6,
	["1BS"] = 0xD5,
	["4BS"] = 0xA5
}

-- { offset, size }
local LOCATIONS = {
	["RPS_R1"]           = {  0,  3 },
	["RPS_EB"]           = {  3,  1 },
	["RPS_T21"]          = {  2,  1 },
	["RPS_NU"]           = {  3,  1 },
	["1BS_LRN"]          = {  4,  1 },
	["1BS_CO"]           = {  7,  1 },
	["4BS_LRN"]          = { 28,  1 },
	["4BS_LRN_TYPE"]     = { 24,  1 },
	["4BS_LRN_EEP_FUNC"] = {  0,  6 },
	["4BS_LRN_EEP_TYPE"] = {  6,  7 },
	["4BS_LRN_MANUF_ID"] = { 13, 11 },
	["4BS_02_TMP8"]      = { 16,  8 },
	["4BS_02_TMP10"]     = { 14, 10 },
	["4BS_04_HUM"]       = {  8,  8 },
	["4BS_04_TMP"]       = { 16,  8 },
	["4BS_04_TSN"]       = { 30,  1 },
	["4BS_06_RS"]        = { 31,  1 },
	["4BS_06_ILL1"]      = { 16,  8 },
	["4BS_06_ILL2"]      = {  8,  8 },
	["4BS_07_PIRS"]      = { 16,  8 }
}

local ROCKER_BUTTONS = {
	[0] = "AI",
	[1] = "AO",
	[2] = "BI",
	[3] = "BO",
	[4] = "CI",
	[5] = "CO",
	[6] = "DI",
	[7] = "DO",
}

local CRC8_TABLE = {
	0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15,
	0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
	0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65,
	0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
	0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5,
	0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
	0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85,
	0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
	0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
	0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
	0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2,
	0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
	0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32,
	0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
	0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42,
	0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
	0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c,
	0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
	0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec,
	0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
	0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c,
	0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
	0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c,
	0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
	0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b,
	0x76, 0x71, 0x78, 0x7f, 0x6A, 0x6d, 0x64, 0x63,
	0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
	0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
	0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb,
	0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8D, 0x84, 0x83,
	0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb,
	0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
}

------------------------------------------------------------------------------------------------------------------------
-- Globals
------------------------------------------------------------------------------------------------------------------------

local g_parentDevice        -- The device # of the parent device
local g_ptr                 -- Pointer used to create devices

local g_taskHandle = -1     -- Handle for the system messages
local g_lastSysMessage = 0  -- Timestamp of the last status message

local g_messageQueue = {}   -- The outbound message queue

local g_childDevices = {}   -- The list of all our child devices
local g_MaxId = 0           -- A number that increments with every device learned.

local g_knownSenderIds = {} -- The list of the newly learned EnOcean devices while in Learn Mode.
local g_newDevices = {      -- The list of the newly appended Luup devices while in Learn Mode.
	-- altid
	-- description
	-- deviceType
	-- deviceFile
	-- implementationFile
	-- parameters
}

------------------------------------------------------------------------------------------------------------------------
-- Generic utilities
------------------------------------------------------------------------------------------------------------------------

local log      = luup.log
-- Get the 'getinfo' function from the 'debug' module before overwriting the 'debug' keyword.
local getinfo  = debug.getinfo
local function debug() end


local function ShowSysMessage (message, mode, critical)
	mode = mode or SYS_MESSAGE_TYPES.BUSY
	critical = critical or false
	log( "(EnOceanPlugin::ShowSysMessage-"..(getinfo(2).name or "N/A")..") mode: ".. mode
		..", critical: ".. tostring( critical ) ..", message: ".. message )

	luup.task( message, mode, "EnOcean Plugin", g_taskHandle )
	g_lastSysMessage = tostring( os.time() )

	if not critical then
		-- Clear the previous system message, since it's transient.
		luup.call_delay( "ClearSysMessage", 30, g_lastSysMessage )
	else
		luup.set_failure( true, g_parentDevice )
	end
end


function ClearSysMessage (messageTime)
	-- 'messageTime' is nil if the function is called by the user.
	if (messageTime == g_lastSysMessage) or (messageTime == nil) then
		luup.task( "Clearing...", SYS_MESSAGE_TYPES.SUCCESS, "EnOcean Plugin", g_taskHandle )
	end
end


local function PadLeft (s, length, c)
	s = tostring(s)
	length = length or 2
	c = c or "0"
	return c:rep( length - #s ) .. s
end


-- Formats a number or character as hex.
local function ToHex (n)
	if type(n) == "number" then
		return string.format( "%02X", n )
	end
end


-- Check if a table contains the given item.
-- Returns true and the key/index of the item if found,
-- or false if not found.
local function TableContains (t, item)
	for k, v in pairs(t) do
		if v == item then
			return true, k
		end
	end
	return false
end


-- Extracts a subtable from the given table.
local function ExtractSubtable (t, start, length)
	if start < 0 then
		start = #t + start + 1
	end
	length = length or (#t - start + 1)

	local t1 = {}
	for i = start, start + length - 1 do
		t1[#t1 + 1] = t[i]
	end
	return t1
end


-- Concatenates a table of numbers into a string separated by the given separator.
local function ConcatTableHex( t, sep, start, length )
	sep = sep or "-"
	start = start or 1
	if start < 0 then
		start = #t + start + 1
	end
	length = length or (#t - start + 1)
	local s = ToHex( t[start] )
	if length > 1 then
		for i = start + 1, start + length - 1 do
			s = s .. sep .. ToHex( t[i] )
		end
	end
	return s
end


-- Set "class".
local set = {
	add = function (s, item)
		if not TableContains( s, item ) then
			table.insert( s, item )
		end
	end,
	remove = function (s, item)
		local itemExists, pos = TableContains( t, item )
		if itemExists then
			table.remove( t, pos )
		end
	end,
	concat = table.concat
}


------------------------------------------------------------------------------------------------------------------------
-- Misc plugin functions
------------------------------------------------------------------------------------------------------------------------

local function ComputeCrc (data, start, length)
	start = start or 1
	length = length or #data

	local crc = 0
	for i = start, start + length - 1 do
		crc = CRC8_TABLE[bit.bxor( crc, data[i] ) + 1]
	end
	debug( "(EnOceanPlugin::ComputeCrc) crc=".. ToHex( crc ))
	return crc
end


-- Returns the numerical value made of the 'size' number of bits starting at 'offset'.
-- The offset starts from the last bit of the last byte. The bit at the smallest offset is the MSB.
-- The data is a table where the first byte (according to the EnOcean docs) is on the last position in the table.
-- So, for 4 bytes: data[1]=byte3, data[2]=byte2, data[3]=byte1, data[4]=byte0
local function GetValueAtLocation (data, location)
	local offset = location[1]
	local size   = location[2]
	local value  = 0 -- The return value

	local startingOffset = offset
	local maxOffset = #data * 8 - 1

	local inverseOffset
	local byteIndex
	local bitIndex
	for offset = startingOffset, startingOffset + size - 1 do
		inverseOffset = maxOffset - offset
		byteIndex = math.floor( offset / 8 ) + 1
		bitIndex = inverseOffset % 8
		value = bit.lshift( value, 1 )
		if bit.check( data[byteIndex], 2 ^ bitIndex ) then
			value = bit.set( value, 1 )
		end
	end

	return value
end


local function IsLearned (senderId)
	senderId = senderId:gsub( "%-", "%%%-" ) -- Escape '-' because it's a magic character.
	for altid in pairs( g_childDevices ) do
		if altid:find( "^".. senderId ) then
			return true
		end
	end
	return false
end


local function GetEepFuncAndType (senderId)
	senderId = senderId:gsub( "%-", "%%%-" ) -- Escape '-' because it's a magic character.
	for altid in pairs( g_childDevices ) do
		if altid:find( "^".. senderId ) then
			local eepFunc, eepType = altid:match( "^".. senderId .."%-(%x%x)%-(%x%x)" )
			if eepFunc and eepType then
				return tonumber( eepFunc, 16 ), tonumber( eepType, 16 )
			end
		end
	end
end


------------------------------------------------------------------------------------------------------------------------
-- Radio telegram handlers
------------------------------------------------------------------------------------------------------------------------

local RADIO_TELEGRAM_HANDLERS = {

	--------------------------------------------------------------------------------------------------------------------
	-- RPS
	--------------------------------------------------------------------------------------------------------------------
	[TELEGRAM_TYPES["RPS"]] = function (senderId, data, status)
		debug( "(EnOceanPlugin::RPS) RPS telegram handler" )

		local firstAction = GetValueAtLocation( data, LOCATIONS["RPS_R1"] )
		local altid = senderId ..'-'.. ROCKER_BUTTONS[firstAction]:sub( 1, 1 )
		debug( "(EnOceanPlugin::RPS) altid: ".. altid )

		--------------------------------------------------------------------------------------------
		-- The device is learned and it's not a "number of buttons pressed" telegram.
		--------------------------------------------------------------------------------------------
		if g_childDevices[altid] then
			log( "(EnOceanPlugin::RPS) Device known: ".. g_childDevices[altid] )
			-- Ignore "number of buttons pressed" telegrams.
			if GetValueAtLocation( { status }, LOCATIONS["RPS_NU"] ) == 0 then
				debug( "(EnOceanPlugin::RPS) 'Number of buttons pressed' telegram. Return" )
				return
			end
			-- Ignore "button released" telegrams.
			if GetValueAtLocation( data, LOCATIONS["RPS_EB"] ) == 0 then
				debug( "(EnOceanPlugin::RPS) 'Button released' telegram. Return" )
				return
			end
			local pressed = ( ROCKER_BUTTONS[firstAction]:sub( 2, 2 ) == "I" ) and "1" or "0"
			luup.variable_set( SID.SW_POWER, "Status", pressed, g_childDevices[altid] )

		--------------------------------------------------------------------------------------------
		-- The device is not learned, and we're in Learn Mode.
		-- RPS devices don't have Teach-in telegrams.
		--------------------------------------------------------------------------------------------
		elseif F_LEARN_MODE and not TableContains( g_knownSenderIds, senderId ) then
			-- We don't want to add the same device twice.
			set.add( g_knownSenderIds, senderId )
			g_MaxId = g_MaxId + 1

			-- Find the device type. If T21 is 1, this is a 2 rocker switch, otherwise it's a 4 rocker switch.
			local t21 = GetValueAtLocation( { status }, LOCATIONS["RPS_T21"] )
			local numRockers = (t21 == 1) and 2 or 4

			-- Create the devices.
			log( "(EnOceanPlugin::RPS) T21: ".. t21 ..", num rockers: ".. numRockers )
			ShowSysMessage( "Found ".. numRockers .." Rocker Switch (EO-".. PadLeft( g_MaxId ) ..")" )
			-- Create a switch for each rocker.
			local deviceName
			local parameters = SID.SW_POWER ..",Status=0"
			for i = 0, numRockers - 1 do
				-- Get only the button, ignore the state.
				local button = ROCKER_BUTTONS[2 * i]:sub( 1, 1 )
				altid = senderId ..'-'.. button
				deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Button ".. button
				log(  "(EnOceanPlugin::RPS) New device: ".. deviceName  )
				table.insert( g_newDevices, {
					["altid"]      = altid,
					["deviceName"] = deviceName,
					["deviceType"] = DEVICE_TYPES.BINARY_LIGHT,
					["deviceFile"] = DEVICE_FILES.BINARY_LIGHT,
					["implementationFile"] = "",
					["parameters"] = parameters
				} )
			end
		else
			log( "(EnOceanPlugin::RPS) Unkown device" )
		end
	end,

	--------------------------------------------------------------------------------------------------------------------
	-- 1BS
	--------------------------------------------------------------------------------------------------------------------
	[TELEGRAM_TYPES["1BS"]] = function (senderId, data, status)
		debug(  "(EnOceanPlugin::1BS) 1BS telegram handler"  )

		local isDataTelegram = GetValueAtLocation( data, LOCATIONS["1BS_LRN"] ) == 1
		debug(  "(EnOceanPlugin::1BS) ".. (isDataTelegram and "Data" or "Teach-in") .." telegram"  )

		--------------------------------------------------------------------------------------------
		-- The device is learned and this is a Data telegram.
		--------------------------------------------------------------------------------------------
		if g_childDevices[senderId] and isDataTelegram then
			log(  "(EnOceanPlugin::1BS) Device known: ".. g_childDevices[senderId]  )
			local tripped = 1 - GetValueAtLocation( data, LOCATIONS["1BS_CO"] )
			luup.variable_set( SID.SECURITY, "Tripped", tripped, g_childDevices[senderId] )

		--------------------------------------------------------------------------------------------
		-- The device is not learned, and we're in Learn Mode.
		-- Only learn the device if this is a Teach-in telegram, even if it doesn't contain any useful information.
		-- We don't want to accidentaly learn devices that the user doesn't want.
		--------------------------------------------------------------------------------------------
		elseif F_LEARN_MODE and not isDataTelegram and not g_childDevices[senderId] and not TableContains( g_knownSenderIds, senderId ) then
			-- We don't want to add the same device twice.
			set.add( g_knownSenderIds, senderId )
			g_MaxId = g_MaxId + 1
			log( "(EnOceanPlugin::1BS) New device: ".. deviceName  )
			ShowSysMessage( "Found Single Input Contact (EO-".. PadLeft( g_MaxId ) ..")" )
			deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Single Input Contact"
			table.insert( g_newDevices, {
				["altid"]      = senderId,
				["deviceName"] = deviceName,
				["deviceType"] = DEVICE_TYPES.DOOR_SENSOR,
				["deviceFile"] = DEVICE_FILES.DOOR_SENSOR,
				["implementationFile"] = "",
				["parameters"] = SID.SECURITY ..",Tripped=0\n".. SID.SECURITY ..",Armed=0"
			} )
		else
			log( "(EnOceanPlugin::1BS) Unkown device or Teach-in telegram" )
		end
	end,

	--------------------------------------------------------------------------------------------------------------------
	-- 4BS
	--------------------------------------------------------------------------------------------------------------------
	[TELEGRAM_TYPES["4BS"]] = function (senderId, data, status)
		debug(  "(EnOceanPlugin::4BS) 4BS telegram handler"  )

		local isDataTelegram = GetValueAtLocation( data, LOCATIONS["4BS_LRN"] ) == 1
		debug(  "(EnOceanPlugin::4BS) ".. (isDataTelegram and "Data" or "Teach-in") .." telegram"  )

		--------------------------------------------------------------------------------------------
		-- The device is learned and it's handled, and this is a Data telegram.
		--------------------------------------------------------------------------------------------
		if IsLearned( senderId ) and isDataTelegram then
			local eepFunc, eepType = GetEepFuncAndType( senderId )
			local altid = senderId .."-".. ToHex( eepFunc ) .."-".. ToHex( eepType )
			debug( "(EnOceanPlugin::4BS) altid: ".. altid )
			--
			-- Temperature sensor
			--
			if eepFunc == 0x02 then
				if not g_childDevices[altid] then
					log( "(EnOceanPlugin::4BS) Device not known. Return" )
					return
				end
				log( "(EnOceanPlugin::4BS) Device known: ".. g_childDevices[altid] )

				local sensorType = math.floor( eepType / 0x10 )
				-- Sensor types:
				-- 0 :  8 bit : 40 degrees range
				-- 1 :  8 bit : 80 degrees range
				-- 2 : 10 bit : -10...+41.2 degrees
				-- 3 : 10 bit : -40...+62.3 degrees
				local temp
				if sensorType == 0 then
					local minTemp = -50 + 10 * (eepType % 0x10)
					local value = GetValueAtLocation( data, LOCATIONS["4BS_02_TMP8"] ) -- 255...0
					local temp = math.floor( minTemp + (255 - value) * 40 / 255 ) -- minTemp..minTemp+40
				elseif sensorType == 1 then
					local minTemp = -60 + 10 * (eepType % 0x10)
					local value = GetValueAtLocation( data, LOCATIONS["4BS_02_TMP8"] ) -- 255...0
					local temp = math.floor( minTemp + (255 - value) * 80 / 255 ) -- minTemp...minTemp+80
				elseif sensorType == 2 then
					local minTemp = -10
					local maxTemp = 41.2
					local value = GetValueAtLocation( data, LOCATIONS["4BS_02_TMP10"] ) -- 1023...0
					local temp = math.floor( minTemp + (1023 - value) * (maxTemp - minTemp) / 1023 ) -- minTemp...maxTemp
				elseif sensorType == 3 then
					local minTemp = -40
					local maxTemp = 62.3
					local value = GetValueAtLocation( data, LOCATIONS["4BS_02_TMP10"] ) -- 1023...0
					local temp = math.floor( minTemp + (1023 - value) * (maxTemp - minTemp) / 1023 ) -- minTemp...maxTemp
				end
				if temp then
					luup.variable_set( SID.TEMP, "CurrentTemperature", temp, g_childDevices[altid] )
				end
			--
			-- Humidity and temperature sensors
			--
			elseif eepFunc == 0x04 then
				local altidHum = altid .."-hum"
				local altidTemp = altid .."-temp"
				if not g_childDevices[altidHum] then
					log( "(EnOceanPlugin::4BS) Device not known. Return" )
					return
				end
				log( "(EnOceanPlugin::4BS) Device known: hum ".. g_childDevices[altidHum] ..", temp "..( g_childDevices[altidTemp] or "NONE" ))

				-- Get the humidity.
				local hum = GetValueAtLocation( data, LOCATIONS["4BS_04_HUM"] ) -- 0...250
				hum = math.floor( hum * 100 / 250 ) -- 0...100
				luup.variable_set( SID.HUMIDITY, "CurrentLevel", hum, g_childDevices[altidHum] )
				-- Check if the temperature sensor is available, and get the temperature if it is.
				local tempSensorAvailable = GetValueAtLocation( data, LOCATIONS["4BS_04_TSN"] )
				if tempSensorAvailable == 1 then
					local temp = GetValueAtLocation( data, LOCATIONS["4BS_04_TMP"] ) -- 0...250
					temp = math.floor( temp * 40 / 250 ) -- 0...40
					luup.variable_set( SID.TEMP, "CurrentTemperature", temp, g_childDevices[altidTemp] )
				end
			--
			-- Light sensor
			--
			elseif eepFunc == 0x06 then
				if not g_childDevices[altid] then
					log( "(EnOceanPlugin::4BS) Device not known. Return" )
					return
				end
				log( "(EnOceanPlugin::4BS) Device known: ".. g_childDevices[altid] )

				local range = GetValueAtLocation( data, LOCATIONS["4BS_06_RS"] )
				local location = (range == 0) and LOCATIONS["4BS_06_ILL1"] or LOCATIONS["4BS_06_ILL2"]
				local lux = GetValueAtLocation( data, location ) -- 0...255
				-- Light sensor types:
				-- 1 : 600...60000 / 300...30000 lux
				-- 2 :   0... 1020 /   0...  510 lux
				local minLux = 0
				local maxLux = 0
				if eepType == 0x01 then
					minLux = (range == 0) and 600 or 300
					maxLux = (range == 0) and 60000 or 30000
				elseif eepType == 0x02 then
					maxLux = (range == 0) and 1020 or 510
				end
				lux = math.floor( minLux + lux * (maxLux - minLux) / 255 ) -- minLux...maxLux
				luup.variable_set( SID.TEMP, "CurrentLevel", lux, g_childDevices[altid] )
			--
			-- Occupancy sensor
			--
			elseif eepFunc == 0x07 then
				if not g_childDevices[altid] then
					log( "(EnOceanPlugin::4BS) Device not known. Return" )
					return
				end
				log( "(EnOceanPlugin::4BS) Device known: ".. g_childDevices[altid] )

				local tripped = GetValueAtLocation( data, LOCATIONS["4BS_07_PIRS"] ) -- 0...255
				tripped = (tripped > 127) and 1 or 0
				luup.variable_set( SID.SECURITY, "Tripped", tripped, g_childDevices[altid] )
			end
		--------------------------------------------------------------------------------------------
		-- The device is not learned, we're in Learn Mode, and we received a Teach-in telegram.
		--------------------------------------------------------------------------------------------
		elseif F_LEARN_MODE and not isDataTelegram and not g_childDevices[senderId] and not TableContains( g_knownSenderIds, senderId ) then
			local lrnType = GetValueAtLocation( data, LOCATIONS["4BS_LRN_TYPE"] )
			if lrnType ~= 1 and lrnType ~= 2 then
				log( "(EnOceanPlugin::4BS) Teach-in telegram variation 1, discard it" )
				return
			end

			-- We don't want to add the same device twice.
			set.add( g_knownSenderIds, senderId )
			g_MaxId = g_MaxId + 1

			local eepFunc = GetValueAtLocation( data, LOCATIONS["4BS_LRN_EEP_FUNC"] )
			local eepType = GetValueAtLocation( data, LOCATIONS["4BS_LRN_EEP_TYPE"] )
			local manufId = GetValueAtLocation( data, LOCATIONS["4BS_LRN_MANUF_ID"] )

			log( "(EnOceanPlugin::4BS) Func ".. ToHex( eepFunc ) ..", Type ".. ToHex( eepType ))
			senderId = senderId .."-".. ToHex( eepFunc ) .."-".. ToHex( eepType )

			--
			-- Temperature sensor
			--
			if eepFunc == 0x02 then
				local deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Temperature Sensor"
				log( "(EnOceanPlugin::4BS) New device: ".. deviceName )
				ShowSysMessage( "Found Temperature sensor (EO-".. PadLeft( g_MaxId ) ..")" )
				table.insert( g_newDevices, {
					["altid"]      = senderId,
					["deviceName"] = deviceName,
					["deviceType"] = DEVICE_TYPES.TEMPERATURE_SENSOR,
					["deviceFile"] = DEVICE_FILES.TEMPERATURE_SENSOR,
					["implementationFile"] = "",
					["parameters"] = SID.TEMP ..",CurrentTemperature=0"
				} )
			--
			-- Temperature and humidity sensors
			--
			elseif eepFunc == 0x04 then
				ShowSysMessage( "Found Temperature and Humidity sensor (EO-".. PadLeft( g_MaxId ) ..")" )
				-- Create the humidity sensor.
				deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Humidity Sensor"
				log( "(EnOceanPlugin::4BS) New device: ".. deviceName )
				table.insert( g_newDevices, {
					["altid"]      = senderId .."-hum",
					["deviceName"] = deviceName,
					["deviceType"] = DEVICE_TYPES.HUMIDITY_SENSOR,
					["deviceFile"] = DEVICE_FILES.HUMIDITY_SENSOR,
					["implementationFile"] = "",
					["parameters"] = SID.HUMIDITY ..",CurrentLevel=0"
				} )
				-- Create the temperature sensor.
				local deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Temperature Sensor"
				log( "(EnOceanPlugin::4BS) New device: ".. deviceName )
				table.insert( g_newDevices, {
					["altid"]      = senderId .."-temp",
					["deviceName"] = deviceName,
					["deviceType"] = DEVICE_TYPES.TEMPERATURE_SENSOR,
					["deviceFile"] = DEVICE_FILES.TEMPERATURE_SENSOR,
					["implementationFile"] = "",
					["parameters"] = SID.TEMP ..",CurrentTemperature=0"
				} )
			--
			-- Light sensor
			--
			elseif eepFunc == 0x06 then
				local deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Light Sensor"
				log( "(EnOceanPlugin::4BS) New device: ".. deviceName )
				ShowSysMessage( "Found Light sensor (EO-".. PadLeft( g_MaxId ) ..")" )
				table.insert( g_newDevices, {
					["altid"]      = senderId,
					["deviceName"] = deviceName,
					["deviceType"] = DEVICE_TYPES.LIGHT_SENSOR,
					["deviceFile"] = DEVICE_FILES.LIGHT_SENSOR,
					["implementationFile"] = "",
					["parameters"] = SID.LIGHT ..",CurrentLevel=0"
				} )
			--
			-- Occupancy sensor
			--
			elseif eepFunc == 0x07 then
				local deviceName = "(EO-".. PadLeft( g_MaxId ) ..") Occupancy Sensor"
				log( "(EnOceanPlugin::4BS) New device: ".. deviceName )
				ShowSysMessage( "Found Occupancy sensor (EO-".. PadLeft( g_MaxId ) ..")" )
				table.insert( g_newDevices, {
					["altid"]      = senderId,
					["deviceName"] = deviceName,
					["deviceType"] = DEVICE_TYPES.MOTION_SENSOR,
					["deviceFile"] = DEVICE_FILES.MOTION_SENSOR,
					["implementationFile"] = "",
					["parameters"] = SID.SECURITY ..",Tripped=0\n".. SID.SECURITY ..",Armed=0"
				} )
			else
				log( "(EnOceanPlugin::4BS) Found unknown device" )
				ShowSysMessage( "Found unknown device A5-".. ToHex( eepFunc ) .."-".. ToHex( eepType ) )
			end
		else
			log( "(EnOceanPlugin::4BS) Unkown device or Teach-in telegram" )
		end
	end
}


------------------------------------------------------------------------------------------------------------------------
-- Packet handler functions
------------------------------------------------------------------------------------------------------------------------

local PACKET_HANDLERS = {

	--------------------------------------------------------------------------------------------------------------------
	-- RADIO
	--------------------------------------------------------------------------------------------------------------------
	[PACKET_TYPES.RADIO] = function (data, optData)
		local radioTelegramType = data[1]
		local senderId = ConcatTableHex( data, '-', -5, 4 )
		local status = data[#data]
		data = ExtractSubtable( data, 2, #data - 6 ) -- Get only the user data.
		log( "(EnOceanPlugin::RADIO) Handle radio telegram type: ".. ToHex( radioTelegramType )
			.. ", senderId: ".. senderId
			.. ", status: ".. ToHex( status )
			.. ", data: ".. ConcatTableHex( data, '-' ))

		if RADIO_TELEGRAM_HANDLERS[radioTelegramType] then
			RADIO_TELEGRAM_HANDLERS[radioTelegramType]( senderId, data, status )
		else
			log( "(EnOceanPlugin::RADIO) Unhandled radio telegram type ".. ToHex( radioTelegramType ))
		end
	end,

	--------------------------------------------------------------------------------------------------------------------
	-- RESPONSE
	--------------------------------------------------------------------------------------------------------------------
	[PACKET_TYPES.RESPONSE] = function (data, optData)
		debug( "(EnOceanPlugin::RESPONSE) Handle response not implemented" )
	end
}


------------------------------------------------------------------------------------------------------------------------
-- Handle incoming data
------------------------------------------------------------------------------------------------------------------------

-- 4BS
-- 0x55 | 0x0 0xa 0x7 0x1 | 0xeb | 0xa5 - 0xce 0x79 0x0 0xf - 0x0 0x81 0x1 0x75 - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x44 0x0 | 0x6e
-- 0x55 | 0x0 0xa 0x7 0x1 | 0xeb | 0xa5 - 0xc6 0x79 0x0 0xf - 0x0 0x81 0x1 0x75 - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x43 0x0 | 0x15
-- 0x55 | 0x0 0xa 0x7 0x1 | 0xeb | 0xa5 - 0xc6 0x79 0x0 0xf - 0x0 0x81 0x1 0x75 - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x43 0x0 | 0x15

-- 1BS
-- 0x55 | 0x0 0x7 0x7 0x1 | 0x7a | 0xd5 - 0x8 - 0x0 0x1 0x8a 0xbb - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x3a 0x0 | 0xbc
-- 0x55 | 0x0 0x7 0x7 0x1 | 0x7a | 0xd5 - 0x8 - 0x0 0x0 0x85 0xe - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x37 0x0 | 0x3a

local state        = STATE.GET_SYNC_STATE -- The current state
local rxBuf        = {} -- The received data buffer
local rxCount      = 0  -- The number of bytes in the data buffer
local rxDataLen    = 0  -- The length of the data field
local rxOptDataLen = 0  -- The length of the optional data field
local rxPacketType = 0  -- The packet type

-- State machine to load incoming packet bytes
local STATE_MACHINE = {

	-- Waiting for packet sync byte 0x55
	[STATE.GET_SYNC_STATE] = function (rxByte)
		if rxByte == SYNC_BYTE then
			--debug( "(EnOceanPlugin::GET_SYNC_STATE) Got sync byte, go to GET_HEADER_STATE" )
			state = STATE.GET_HEADER_STATE
			crc   = 0
			rxCount = 0
		end
	end,

	-- Read the header bytes.
	[STATE.GET_HEADER_STATE] = function (rxByte)
		-- Add the received byte in the buffer.
		rxCount = rxCount + 1
		rxBuf[rxCount] = rxByte
		-- All the header bytes received?
		if rxCount == 4 then
			--debug( "(EnOceanPlugin::GET_HEADER_STATE) Received all header bytes, go to CHECK_CRC8H_STATE" )
			state = STATE.CHECK_CRC8H_STATE
		end
	end,

	-- Check the header checksum & try to resynchronize if an error occurred.
	[STATE.CHECK_CRC8H_STATE] = function (rxByte)
		-- Header CRC correct?
        if ComputeCrc( rxBuf, 1, rxCount ) ~= rxByte then
			--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) Incorrect header CRC" )
			-- No. Check if there is a sync byte (0x55) in the header.
			local pos = -1
			for i = 1, 4 do
				if rxBuf[i] == SYNC_BYTE then
					-- Indicates the position of the sync byte
					pos = i
					break
				end
			end

			if pos == -1 and rxByte ~= SYNC_BYTE then
				-- Neither the header nor the CRC8H contain the sync byte
				--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) Neither the header nor the CRC8H contain the sync byte, go to GET_SYNC_STATE" )
				state = STATE.GET_SYNC_STATE
				return
			elseif pos == -1 and rxByte == SYNC_BYTE then
				-- The header does not contain the sync byte but the CRC8H does
				-- The sync byte could be the beginning of a packet
				--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) The header does not contain the sync byte but the CRC8H does, go to GET_HEADER_STATE" )
				state = STATE.GET_HEADER_STATE
				rxCount = 0
				return
			end

			-- The header contains the sync byte. It could be a new packed.
			-- Shift all the bytes from the 0x55 byte in the buffer.
			-- Recalculate CRC8 for those bytes.
			for i = 1, 4 - pos do
				rxBuf[i] = rxBuf[pos + i]
			end
			rxCount = 4 - pos

			-- Copy the received byte to buffer
			rxCount = rxCount + 1
			rxBuf[rxCount] = rxByte

			if rxCount < 4 then
				--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) There are still header bytes to receive, go to GET_HEADER_STATE" )
				state = STATE.GET_HEADER_STATE
			end

			--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) All header bytes received, stay in CHECK_CRC8H_STATE" )
			return
        end

		-- CRC8H correct. Are the Length fields values valid?
		rxDataLen = rxBuf[1] * 0x100 + rxBuf[2]
		rxOptDataLen = rxBuf[3]
		--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) Header CRC correct, rxDataLen=".. rxDataLen ..", rxOptDataLen=".. rxOptDataLen )
		if (rxDataLen + rxOptDataLen) == 0 then
			-- No. Sync byte received?
			if rxByte == SYNC_BYTE then
				-- Yes
				--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) Header CRC incorrect, sync byte received, go to GET_HEADER_STATE" )
				state = STATE.GET_HEADER_STATE
				rxCount = 0
				return
			end

			-- Packet with correct CRC8H, but wrong length fields
			--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) Packet with correct CRC8H, but wrong length fields, go to GET_SYNC_STATE" )
			state = STATE.GET_SYNC_STATE
			return
		end

		-- Correct header CRC8. Go to the reception of data.
		--debug( "(EnOceanPlugin::CHECK_CRC8H_STATE) Correct header CRC8, go to GET_DATA_STATE" )
		state = STATE.GET_DATA_STATE
		rxCount = 0
		rxPacketType = rxBuf[4]
	end,

	-- Copy the information bytes.
	[STATE.GET_DATA_STATE] = function (rxByte)
		rxCount = rxCount + 1
		rxBuf[rxCount] = rxByte

		-- When all the expected bytes are received, go to calculate data checksum.
		if rxCount == (rxDataLen + rxOptDataLen) then
			--debug( "(EnOceanPlugin::GET_DATA_STATE) All expected bytes received, go to CHECK_CRC8D_STATE" )
			state = STATE.CHECK_CRC8D_STATE
		end
	end,

	-- Check the data CRC8
	[STATE.CHECK_CRC8D_STATE] = function (rxByte)
		-- In all cases the state returns to the first state: waiting for next sync byte.
		state = STATE.GET_SYNC_STATE

		if ComputeCrc( rxBuf, 1, rxCount ) == rxByte then
			-- Correct packet received
			--debug( "(EnOceanPlugin::CHECK_CRC8D_STATE) Correct packet received, handle it and go to GET_SYNC_STATE" )
			if PACKET_HANDLERS[rxPacketType] then
				PACKET_HANDLERS[rxPacketType]( ExtractSubtable( rxBuf, 1, rxDataLen ), ExtractSubtable( rxBuf, rxDataLen + 1, rxOptDataLen ))
			else
				log( "(EnOceanPlugin::CHECK_CRC8D_STATE) Unhandled packet type ".. ToHex( rxPacketType ))
			end
			return
		end

		-- Incorrect CRC8
		-- If the received byte is the sync byte, then it could be the sync byte for next packet.
		if rxByte == SYNC_BYTE then
			--debug( "(EnOceanPlugin::CHECK_CRC8D_STATE) Incorrect CRC8, the received byte is the sync byte, go to GET_HEADER_STATE" )
			state = STATE.GET_HEADER_STATE
			rxCount = 0
		end

		log( "(EnOceanPlugin::CHECK_CRC8D_STATE) Error: incorrect data checksum" )
	end
}


function HandleIncoming (lul_data)
	local rxByte = string.byte( lul_data )
	--debug( "(EnOceanPlugin::HandleIncoming) state ".. state ..", rxByte=".. ToHex( rxByte ))
	if STATE_MACHINE[state] then
		STATE_MACHINE[state]( rxByte )
	else
		-- Shouldn't get here.
		-- Wait for the sync byte.
		state = STATE.GET_SYNC_STATE
	end
end


------------------------------------------------------------------------------------------------------------------------
-- Message sending functions
------------------------------------------------------------------------------------------------------------------------

local function AddToSendQueue (packetType, data, optData, waitForResponse)
	local insert = table.insert
	local char   = string.char

	local packet = { SYNC_BYTE }              -- The sync byte

	insert( packet, char( #data / 0x100 ))    -- First byte of the data length
	insert( packet, char( #data % 0x100 ))    -- Seconds byte of the data length
	insert( packet, char( #optData ))         -- Length of the optional data
	insert( packet, packetType )              -- The packet type
	insert( packet, ComputeCrc(packet, 2, 4)) -- The header CRC8

	insert( packet, data )                    -- The user data
	insert( packet, optData )                 -- The optional user data
	insert( packet, ComputeCrc( packet, 7, #data + #optData )) -- The data CRC8

	table.insert( g_messageQueue, {content = table.concat(packet), waitForResponse = (waitForResponse or false)} )
	if not F_SENDING_IN_PROGRESS then
		SendPacket()
	end
end


function SendPacket()
	F_SENDING_IN_PROGRESS = true

	if F_WAIT_FOR_RESPONSE then
		g_triesCount = g_triesCount + 1
		if g_triesCount < 3 then
			debug( "(EnOceanPlugin::SendPacket) F_WAIT_FOR_RESPONSE flag is set, g_triesCount=".. g_triesCount .."; try again in ".. DELAYS.SEND_NEXT_MESSAGE .." seconds." )
			luup.call_delay( "SendPacket", DELAYS.SEND_NEXT_MESSAGE )
			return
		else
			log( "(EnOceanPlugin::SendPacket) Didn't receive response for the last message. Stop waiting for response." )
			F_WAIT_FOR_RESPONSE = false
			g_triesCount = 0
		end
	end

	local message
	if #g_messageQueue > 0 then
		message = g_messageQueue[1]
	else
		return
	end

	debug( "(EnOceanPlugin::SendPacket) Wait for response: ".. message.waitForResponse )

	if not luup.io.write( message.content ) then
		ShowSysMessage( "Failed to send message", SYS_MESSAGE_TYPES.ERROR )
		return
	end

	F_WAIT_FOR_RESPONSE = message.waitForResponse
	table.remove( g_messageQueue, 1 )

	-- Send the next message in the queue.
	if #g_messageQueue > 0 then
		luup.call_delay( "SendPacket", DELAYS.SEND_NEXT_MESSAGE )
	else
		F_SENDING_IN_PROGRESS = false
	end
end


------------------------------------------------------------------------------------------------------------------------
-- Action implementations
------------------------------------------------------------------------------------------------------------------------

function SetLearnMode (learnMode)
	luup.variable_set( SID.ENOCEAN, "LearnMode", learnMode, g_parentDevice )
	F_LEARN_MODE = (learnMode == "1")
	-- We entered Learn Mode. Display a message and do nothing.
	if F_LEARN_MODE then
		ShowSysMessage( "Entered Learn Mode" )
	-- We exited Learn Mode.
	-- If we learned any new device, create its Luup counterpart and update the number of EnOcean devices learned.
	elseif #g_newDevices > 0 then
		luup.variable_set( SID.ENOCEAN, "MaxID", g_MaxId, g_parentDevice )
		local ptr = luup.chdev.start( g_parentDevice )
		for i, v in ipairs( g_newDevices ) do
			luup.chdev.append( g_parentDevice, ptr, v.altid, v.deviceName, v.deviceType,
				v.deviceFile, v.implementationFile, v.parameters, false )
		end
		luup.chdev.sync( g_parentDevice, ptr )
	-- We exited Learn Mode and no new device was learned. Display a message and do nothing.
	else
		ShowSysMessage( "Exited Learn Mode. No new device was learned." )
	end
end


------------------------------------------------------------------------------------------------------------------------
-- Startup functions
------------------------------------------------------------------------------------------------------------------------

local function GetChildDevices()
	-- Get the number of learned EnOcean devices.
	g_MaxId = luup.variable_get( SID.ENOCEAN, "MaxID", g_parentDevice )
	g_MaxId = tonumber( g_MaxId, 10 ) or 0
	-- Get a list with all our child devices.
	for k, v in pairs( luup.devices ) do
		if v.device_num_parent == g_parentDevice then
			g_childDevices[v.id] = k
			debug( "(EnOceanPlugin::GetChildDevices) Found device #".. k .." - ".. v.description ..", sender ID: ".. v.id )
		end
	end
end


------------------------------------------------------------------------------------------------------------------------
-- Init
------------------------------------------------------------------------------------------------------------------------

function Init (lul_device)
	log( "(EnOceanPlugin::Init) EnOcean ESP3 Gateway plugin" )

	-- Get the master device.
	g_parentDevice = lul_device

	-- Get a handle for system messages.
	g_taskHandle = luup.task( "Starting up...", 1, "EnOcean plugin", -1 )

	-- Update the Debug Mode.
	local debugMode = luup.variable_get( SID.ENOCEAN, "DebugMode", lul_device ) or ""
	if debugMode == "" then
		luup.variable_set( SID.ENOCEAN, "DebugMode", (F_DEBUG_MODE and "1" or "0"), lul_device )
	else
		F_DEBUG_MODE = (debugMode == "1") and true or false
	end

	if F_DEBUG_MODE then
		debug = log
	end

	-- Disable Learn Mode.
	luup.variable_set( SID.ENOCEAN, "LearnMode", "0", lul_device )

	GetChildDevices()

	-- Get the sender ID of the TCM300.
	--AddToSendQueue( PACKET_TYPES.COMMON_COMMAND, COMMON_COMMAND.CO_RD_IDBASE, "", true )

	debug( "(EnOceanPlugin::Init) Success: startup successful" )
	return true, "startup successful", "EnOcean plugin"
end
